/*
Copyright (c) 2003-2009, CKSource - Frederico Knabben. All rights reserved.
For licensing, see LICENSE.html or http://ckeditor.com/license
*/
( function()
{
	CKEDITOR.plugins.add( 'pastefromword',
	{
		init : function( editor )
		{
			// Register the command.
			editor.addCommand( 'pastefromword', new CKEDITOR.dialogCommand( 'pastefromword' ) );
			// Register the dialog.
			CKEDITOR.dialog.add( 'pastefromword', this.path + 'dialogs/pastefromword.js' );

			// Register the toolbar button.
			editor.ui.addButton( 'PasteFromWord',
				{
					label : editor.lang.pastefromword.toolbar,
					command : 'pastefromword'
				} );

			editor.on( 'paste', function( evt )
			{
				var data = evt.data,
					mswordHtml;
				// MS-WORD format sniffing.
				if ( ( mswordHtml = data[ 'html' ] )
					 && ( evt.data[ 'ms-word' ]
						  || /(class=\"?Mso|style=\"[^\"]*\bmso\-|w:WordDocument)/.test( mswordHtml )
						     && confirm( editor.lang.pastefromword.confirmCleanup ) ) )
				{
					// Firefox will be confused by those downlevel-revealed IE conditional
					// comments, fixing them first( convert it to upperlevel-revealed one ).
					// e.g. <![if !vml]>...<![endif]>
					if( CKEDITOR.env.gecko )
					{
						data[ 'html' ] =
							mswordHtml.replace( /(<!--\[if[^<]*?\])-->([\S\s]*?)<!--(\[endif\]-->)/gi, '$1$2$3' );
					}

					var filter = data.processor.dataFilter;
					// These rules will have higher priorities than default ones.
					filter.addRules( CKEDITOR.plugins.pastefromword.getRules( editor ), 5 );
				}
			} );
		}
	} );

	CKEDITOR.plugins.pastefromword =
	{
		utils :
		{
			// Create a <cke:listbullet> which indicate an list item type.
			createListBulletMarker : function ( bulletStyle )
			{
				var marker = new CKEDITOR.htmlParser.element( 'cke:listbullet' ),
					listType;

				// TODO: Support more list style type from MS-Word.
				if ( bulletStyle[ 2 ] )
				{
					if ( !isNaN( bulletStyle[ 1 ] ) )
						bulletStyle = 'decimal';
					else if ( bulletStyle[ 1 ].search(/[a-z]/) != -1 )
						bulletStyle = 'lower-latin';
					else if ( bulletStyle[ 1 ].search(/[A-Z]/) != -1 )
						bulletStyle = 'upper-latin';
					else
						bulletStyle = 'decimal';

					listType = 'ol';
				}
				else
				{
					if ( bulletStyle[ 1 ].search(/[l·•]/) != -1 )
						bulletStyle = 'disc';
					else if ( bulletStyle[ 1 ].search(/[oØ]/) != -1 )
						bulletStyle = 'circle';
					else if ( bulletStyle[ 1 ].search(/[n◆]/) != -1 )
						bulletStyle = 'square';
					else
						bulletStyle = 'disc';

					listType = 'ul';
				}

				// Represent list type as CSS style.
				marker.attributes =
				{
					'cke:listtype' : listType,
					'style' : 'list-style-type:' + bulletStyle 
				};

				return marker;
			},

			isListBulletIndicator : function( element )
			{
				var styleText = element.attributes && element.attributes.style;
				if( styleText == 'mso-list: Ignore' )
					return true;
			}

		},

		filters : {
				/**
				 * A simple filter which always rejecting.
				 */
				falsyFilter  : function( value )
				{
					return false;
				},

				/**
				 * A filter dedicated on the 'style' attribute for dropping/replacing style rules.
				 * @param styles {Array} A triple in form of [ styleNameRegexp, styleValueRegexp, newStyleValue ] where the last two are optional.
				 */
				stylesFilter : function( styles )
				{
					return function( styleText, element )
					{
						 var rules = [];
						// html-encoded quote might be introduced by 'font-family'
						// from MS-Word which confused the following regexp. e.g.
						//'font-family: &quot;Lucida, Console&quot;'
						 styleText.replace( /&quot;/g, '"' )
								  .replace( /\s*([^ :;]+?)\s*:\s*([^;]+?)\s*(?=;|$)/g,
							 function( match, name, value )
							 {
								 name = name.toLowerCase();
								 if ( name == 'font-family' )
									value = value.replace( /"/g, '' );

								 var namePattern,
									 valuePattern,
									 newValue,
									 newName;
								 for( var i = 0 ; i < styles.length; i++ )
								 {
									if( styles[ i ] )
									{
										namePattern = styles[ i ][ 0 ];
										valuePattern = styles[ i ][ 1 ];
										newValue = styles[ i ][ 2 ];
										newName = styles[ i ][ 3 ];

										if ( name.match( namePattern )
											 && ( !valuePattern || value.match( valuePattern ) ) )
										{
											name = newName || name;
											if( typeof newValue == 'function' )
												newValue = newValue( value, element );
											if( typeof newValue == 'string' )
												rules.push( [ name, newValue ] );
											return;
										}
									}
								 }
								 rules.push( [ name, value ] );

							 } );

						for ( var i = 0 ; i < rules.length ; i++ )
							 rules[ i ] = rules[ i ].join( ':' );
						return rules.length ?
						         ( rules.join( ';' ) + ';' ) : false;
					 };
				},

				/**
				 * Migrate the element by decorate styles on it.
				 * @param styleDefiniton
				 * @param variables
				 */
				elementMigrateFilter : function ( styleDefiniton, variables )
				{
					return function( element )
					{
						var styleDef =
								variables ?
									new CKEDITOR.style( styleDefiniton, variables )._.definition
									: styleDefiniton;
						element.name = styleDef.element;
						CKEDITOR.tools.extend( element.attributes, CKEDITOR.tools.clone( styleDef.attributes ) );
						var attrs = element.attributes;
						attrs.style = ( attrs.style || '' ) + CKEDITOR.style.getStyleText( styleDef );
					}
				},

				/**
				 * Migrate styles by creating a new nested stylish element.    
				 * @param styleDefinition
				 */
				styleMigrateFilter : function( styleDefinition, variableName )
				{

					var elementMigrateFilter = this.elementMigrateFilter;
					return function( value, element )
					{
						// Build an stylish element first.
						var styleElement = new CKEDITOR.htmlParser.element( null, {} ),
							varialbes = {};

						varialbes[ variableName ] = value;
						elementMigrateFilter( styleDefinition, varialbes )( styleElement );
						// Place the new element inside the existing span.
						styleElement.children = element.children;
						element.children = [ styleElement ];
					};
				},
				
				/**
				 * A filter which remove cke-namespaced-attribute on
				 * all none-cke-namespaced elements. 
				 * @param value
				 * @param element
				 */
				bogusAttrFilter : function( value, element )
				{
					if( element.name.indexOf( 'cke:' ) == -1 )
						return false;
				}

			},

		getRules : function( editor )
		{
			var falsyFilter = this.filters.falsyFilter,
				stylesFilter = this.filters.stylesFilter,
				elementMigrateFilter = this.filters.elementMigrateFilter,
				styleMigrateFilter = CKEDITOR.tools.bind( this.filters.styleMigrateFilter, this.filters ),
				bogusAttrFilter = this.filters.bogusAttrFilter,
				createListBulletMarker = this.utils.createListBulletMarker,
				isListBulletIndicator = this.utils.isListBulletIndicator,
				listDtdParents = CKEDITOR.dtd.parentOf( 'ol' ),
				config = editor.config,
				ignoreFontFace = config.pasteFromWordIgnoreFontFace;

			return {

				elementNames :
				[
					// Remove script, meta and link elements.
					[ /meta|link|script|style/, '' ]
				],

				elements :
				{
					$ : function( element )
					{
						var tagName = element.name || '';

						// Processing headings.
						if ( tagName.match( /h\d/ ) )
						{
							element.filterChildren();
							var onlyChild = element.onlyChild();
							// Remove empty headings.
							if( onlyChild && onlyChild.value
								&& !CKEDITOR.tools.trim( onlyChild.value ) )
								return false;
							delete element.attributes;
							// Migrate heading formatting to editor configured ones.
							elementMigrateFilter( config[ 'format_' + tagName ] )( element );
						}
						// Remove inline elements which contain only empty spaces.
						else if( tagName.match( /^(:?b|u|i|strike|span)$/ ) )
						{
							element.filterChildren();
							var child = element.onlyChild();
							if ( child && /^(:?\s|&nbsp;)+$/.exec( child.value ) )
								delete element.name;
						}
						// Remove namespaced element while preserving the content.
						else if( tagName.indexOf( ':' ) != -1
								 && tagName.indexOf( 'cke' ) == -1 )
						{
							element.filterChildren();

							// Restore img element from vml.
							if( tagName == 'v:imagedata' )
							{
								var href = element.attributes[ 'o:href' ];
								if ( href )
									element.attributes.src = href;
								element.name = 'img';
							}
							else
								delete element.name;
						}
						else if ( !tagName )
						{
							// Trim empty spaces at the beginning of document for IE. 
							if( CKEDITOR.env.ie )
							{
								var firstTextChild = element.firstTextChild();
								if ( firstTextChild
									 && firstTextChild.value.match( /^(:?\s|&nbsp;)+$/ ) )
									element.children.splice( 0, 1 );
							}
						}

						// Any dtd-valid element which could contain a list.
						if( !tagName && element.children
								 || tagName in listDtdParents )
						{
							element.filterChildren();

							var children = element.children, child,
								listItem,   // The current processing cke:li element.
								listItemAttrs,
								listType,   // Determine the root type of the list.
								listItemIndent, // Indent attribute represent the level of it.
								lastListItem, // The previous one just been added to the list.
								list, parentList, // Current staging list and it's parent list if any.  
								indent;

							for( var i = 0 ; i < children.length; i++ )
							{
								child = children[ i ];

								if ( 'cke:li' == child.name )
								{
									child.name = 'li';
									listItem = child;
									listItemAttrs = listItem.attributes;
									listType = listItem.attributes[ 'cke:listtype' ];
									// The indent attribute might not present.
									listItemIndent = listItemAttrs[ 'cke:indent' ] || 0;

									// Ignore the 'list-style-type' attribute if it's matched with
									// the list root element's default style type.
									listItemAttrs.style = stylesFilter(
										[ [ 'list-style-type', listType == 'ol'? 'decimal' : 'disc' ] ] )( listItemAttrs.style ) || '' ;

									if ( !list )
									{
										parentList = list = new CKEDITOR.htmlParser.element( listType );
										list.add( listItem );
										children[ i ] = list;
									}
									else
									{
										if( listItemIndent > indent )
										{
											parentList = list;
											list = new CKEDITOR.htmlParser.element( listType  );
											list.add( listItem );
											lastListItem.add( list );
										}
										else if( listItemIndent < indent )
										{
											list = parentList;
											parentList = list.parent ? list.parent.parent : list;
											list.add( listItem );
										}
										else
											list.add( listItem );

										children.splice( i-- , 1 );
									}

									lastListItem = listItem;
									indent = listItemIndent;
								}
								else
									list = null;
							}

							// Filter childrens again for cleaning up
							// the list attributes.
							element.filterChildren();
						}
					},

					'p' : function( element )
					{
						element.filterChildren();

						var attrs = element.attributes,
							parent = element.parent,
							children = element.children,
							firstChild = children && children[ 0 ];

						// Drop the single wrapper paragraph within table cell
						// while preserve the styles.
						if( /td|th/.test( parent.name )
							&& parent.onlyChild() )
						{
							if( attrs && attrs.style )
								parent.attributes.style += ( attrs.style + ';' );
							delete element.name;
							return;
						}
						// <cke:listbullet> lies inside any paragraph indicate a list item.
						var childs;
						if( ( childs = element.anyChildWithName( 'cke:listbullet' ) )
							 && childs.length )
						{
							var listMaker = childs[ 0 ];
							element.name = 'cke:li';
							attrs.style = stylesFilter(
							[
								[ 'text-indent' ],
								[ 'margin-left', null, function( value )
								{
									// Resolve indent level from 'margin-left' style.
									attrs[ 'cke:indent' ] = parseInt( value );
								} ]
							] )( attrs.style, element ) || '' ;

							// Inherit list-type-style from bullet. 
							var listBulletAttrs = listMaker.attributes,
								listBulletStyle = listBulletAttrs.style;
							attrs.style += ( listBulletStyle + ';' );
							CKEDITOR.tools.extend( attrs, listBulletAttrs );
							children.splice( 0, 1 );
							return;
						}

						// Migrate paragraph formatting based on editor's enter-mode.
						if( config.enterMode == CKEDITOR.ENTER_BR )
						{
							delete element.name;
							element.add( new CKEDITOR.htmlParser.element( 'br' ) );
						}
						else
							elementMigrateFilter( config[ 'format_' + ( config.enterMode == CKEDITOR.ENTER_P ? 'p' : 'div' ) ] )( element );
					},

					'div' : elementMigrateFilter( config[ 'format_' + ( config.enterMode == CKEDITOR.ENTER_P ? 'p' : 'div' ) ] ),
					
					// Deprecates <font> in favor of stylish <span>.
					'font' : function( element )
					{
						// IE: drop the font tag if it comes from list bullet text. 
						if ( CKEDITOR.env.ie && isListBulletIndicator( element.parent ) )
						{
							delete element.name;
							return;
						}

						element.filterChildren();
						// Merge nested <font> tags.
						var parent = element.parent;
						if( 'font' == parent.name )
						{
							CKEDITOR.tools.extend( parent.attributes,
									element.attributes );
							delete element.name;
							return;
						}
						// Convert the topmost into a span with font-styles.
						else
						{
							var attrs = element.attributes,
								styleText = '';
							if( attrs.color )
								styleText += 'color:' + attrs.color + ';';
							if( attrs.face )
								styleText += 'font-family:' + attrs.face + ';';
							// TODO: Mapping size in ranges of xx-small,
							// x-small, small, medium, large, x-large, xx-large.
							if( attrs.size )
								styleText += 'font-size:' +
								            ( attrs.size > 3 ? 'larger'
										      : ( attrs.size < 3 ? 'smaller' : 'medium' ) )+ ';';

							if( styleText )
								element.attributes = { 'style' : styleText };

							element.name = 'span';
						}
					},

					'span' : function( element )
					{
						// IE: remove the span if it comes from list bullet text.
						if ( CKEDITOR.env.ie && isListBulletIndicator( element.parent ) )
							return false;

						element.filterChildren();

						// List item bullet type is supposed to be indicated by
						// the text of a span with style 'mso-list : Ignore'.
						if ( CKEDITOR.env.ie && isListBulletIndicator( element ) )
							return createListBulletMarker( element.firstTextChild().value.match( /([^\s])([.)]?)/ ) );
						
						// Update the src attribute of image element with href.
						var children = element.children,
							firstChild = children && children[ 0 ],
							secondChild;
						if( firstChild
							&& 'cke:imagesource' == firstChild.name )
						{
							secondChild = children[ 1 ];
							if ( 'img' == secondChild.name )
								secondChild.attributes.src = firstChild.attributes.src;
							children.splice( 0, 1 );
							delete element.name;
						}

						// Migrate font related styles.
						var attrs = element.attributes,
							styleText = attrs && attrs.style;
						if( styleText )
						attrs.style = stylesFilter(
									[
										[ 'font-family', null, styleMigrateFilter( config[ 'font_style' ], 'family' ) ],
										// TODO: Convert 'pt' length unit into 'px'.
										[ 'font-size', null, styleMigrateFilter( config[ 'fontSize_style' ], 'size' ) ],
										// TODO: Convert 'rgb' and 'descriptive' color into 'hexadecimal'. 
										[ /^color$/, null, styleMigrateFilter( config[ 'colorButton_foreStyle' ], 'color' ) ],
										[ 'background-color', null, styleMigrateFilter( config[ 'colorButton_backStyle' ], 'color' ) ]
									] )( styleText, element ) || '';
					},

					// Migrate basic style formats to editor configured ones.
					'b' : elementMigrateFilter( config[ 'coreStyles_bold' ] ),
					'i' : elementMigrateFilter( config[ 'coreStyles_italic' ] ),
					'u' : elementMigrateFilter( config[ 'coreStyles_underline' ] ),
					's' : elementMigrateFilter( config[ 'coreStyles_strike' ] ),
					'sup' : elementMigrateFilter( config[ 'coreStyles_superscript' ] ),
					'sub' : elementMigrateFilter( config[ 'coreStyles_subscript' ] ),
					'v:imagedata' : function( element )
					{
						var href = element.attributes['o:href'];
						if( href )
							element.attributes.src = href;
						element.name = 'img';
					}
				},

				attributeNames :
				[
					// Remove onmouseover and onmouseout events (from MS Word comments effect)
					[ /^onmouse(:?out|over)/, '' ],
					// Remove office and vml attribute from elements.
					[ /(?:v|o):\w+/, '' ],
					// Remove lang/language attributes.
					[ /^lang/, '' ],
					ignoreFontFace ? [ /^(?:face|font|size)/, '' ] : null
				],

				attributes :
				{
					'cke:listtype' : bogusAttrFilter,
					'cke:indent' : bogusAttrFilter,
					// Remove mso-xxx styles.
					// Remove margin styles.
					'style' : stylesFilter(
					[
						[ /mso-/ ],
						[ /-moz-/ ],
						// Replace verbose background color style.
						[ /^background$/, null, function( value )
						{
							return value.match( /^[^\s]+/ )[ 0 ];
						}, 'background-color' ],
						[ 'background-color', 'transparent' ],
						// Remove verbose border-color style within Firefox.
						CKEDITOR.env.gecko ? [ 'border-color', /(:?windowtext|-moz-use-text-color|\s)*/ ] : null,
						// 'Indent' format migration(to use editor's indent unit).
						[ /margin-?/, null, function( value, element )
						{
							if( element.name == 'p' )
							{
								value = value.replace( /\d*\.?\d+pt/g, function( length )
								{
									var pt = parseInt( length );
									// Assume MS-Word indent unit length as '11pt'.
									return ( pt / 11 * config.indentOffset ) + config.indentUnit;
								} );
							}
							return value;
						} ],
						[ 'margin', /0(?:cm|in) 0(?:cm|in) 0pt/ ],
						[ 'text-indent', '0cm' ],
						[ 'page-break-before' ],
						[ 'tab-stops' ],
						[ 'display', 'none' ],
						[ 'text-align', 'left' ],
						ignoreFontFace ? [ /font-?/ ] : null,
					] ),
					'width' : function( value, element )
					{
						// Prefer width style over attribute on table cell.
						if( element.name in { td : 1, th : 1 } )
							return false;
					},
					
					'class' : function( value )
					{
						if( value.match( /^(:?Mso|Spell|ListParagraph)/i ) )
							return false;
					},

					// We always have both 'text-align' style company with
					// 'align' attribute, drop the attribute.
					'align' : falsyFilter
				},

				// Fore none-IE, some useful data might be buried under these IE-conditional
				// comments where RegExp were the right approach to dig them out where usual approach
				// is transform it into a fake element node which hold the desired data.
				comment : !CKEDITOR.env.ie ? function( value )
				{
					var imageInfo = value.match( /<img.*?>/ ),
						imageSource = value.match( /<v:imagedata[^>]*o:href=['"](.*?)['"]/ ),
						listInfo = value.match( /^\[if !supportLists\]([\s\S]*?)\[endif\]$/ );

					// Reveal the <img> element in conditional comments for Firefox.
					if( CKEDITOR.env.gecko && imageInfo )
						return CKEDITOR.htmlParser.fragment.fromHtml( imageInfo[ 0 ] ).children[ 0 ];
					// Try to dig the real image link from vml markup.
					if( imageSource )
						return new CKEDITOR.htmlParser.element( 'cke:imagesource', { src : imageSource[ 1 ] } );
					// Seek for list bullet style indicator.
					else if ( listInfo )
					{
						var listSymbol = listInfo[ 1 ],
							listType = listSymbol.match( />([^\s])([.)]?)</ );

						return createListBulletMarker( listType );
					}
					return false;
				} : falsyFilter
			};
		}
	}

	var fragmentPrototype = CKEDITOR.htmlParser.fragment.prototype,
		elementPrototype = CKEDITOR.htmlParser.element.prototype;

	fragmentPrototype.onlyChild = elementPrototype.onlyChild = function()
	{
		var children = this.children,
			count = children.length,
			firstChild = ( count == 1 ) && children[ 0 ];
		return firstChild || null;
	};

	elementPrototype.anyChildWithName = function( tagName, includingSelf )
	{
		var children = this.children,
			count = children && children.length,
			childs = [];

		if( includingSelf && this.name == tagName )
			childs.push( this );

		for ( var i = 0; i < count; i++ )
		{
			if( children[ i ].name  )
				childs = childs.concat( children[ i ].anyChildWithName( tagName, true ) );
		}
		return childs;
	};

	fragmentPrototype.firstTextChild = elementPrototype.firstTextChild = function()
	{
		var child;
		for( var i = 0 ; i < this.children.length ; i++ )
		{
			child = this.children[ i ];
			if( child.value )
				return child;
		}
	};

	/**
	 * Return the DTD-valid parent tag names of the specified one.
	 * @param tagName
	 */
	CKEDITOR.dtd.parentOf = function( tagName )
	{
		var result = {};
		for( var tag in this )
		{
			if( tag.indexOf( '$' ) == -1 && this[ tag ][ tagName ] )
				result[ tag ] = 1;
		}
		return result;
	};

} )();


/**
 * Whether the "Ignore font face definitions" checkbox is enabled by default in
 * the Paste from Word dialog.
 * @type Boolean
 * @default true
 * @example
 * config.pasteFromWordIgnoreFontFace = false;
 */
CKEDITOR.config.pasteFromWordIgnoreFontFace = false;
